// @ts-ignore
import parsePropTypes from 'parse-prop-types'
// @ts-ignore
import { Parser, ProcessingInstructions } from 'html-to-react'
import { RefObject, ReactElement, ReactNode } from 'react'
import { Root } from 'react-dom/client'

const toCamelCase = (str: string = ''): string => {
  return str.replace(/-(\w)/g, (_, c) => (c ? c.toUpperCase() : ''))
}

const toDashedStyle = (str: string = ''): string => {
  return str.replace(/([a-z0-9])([A-Z])/g, '$1-$2').toLowerCase()
}

const handleValueAttr = (children: any[]): any[] => {
  return children.map((child) => {
    if (typeof child.props === 'object') {
      child = { ...child }
      child.props = { ...child.props }
      if (
        ['input', 'textarea', 'select'].indexOf(child.type) >= 0 &&
        child.props.hasOwnProperty('value')
      ) {
        child.props.defaultValue = child.props.value
        delete child.props.value
      }

      if (child.props?.children) {
        let c = Array.isArray(child.props.children)
          ? child.props.children
          : [child.props.children]
        child.props.children = handleValueAttr(c)
      }
    }

    return child
  })
}

const parseChildren = (str: string): ReactNode[] | null => {
  const isValidNode = (node: any): boolean => {
    return !(node.type === 'text' && !node.data.trim())
  }

  const parser = new Parser()
  const processingInstructions = new ProcessingInstructions()
  const child = parser.parseWithInstructions(
    str,
    isValidNode,
    processingInstructions.defaultProcessingInstructions
  )

  if (!child) {
    return null
  }

  let reactChildren =
    child instanceof Array ? child.filter((child) => child !== false) : [child]

  return handleValueAttr(reactChildren)
}

const getPropType = (typeInfo: any): any => {
  if (typeof typeInfo === 'object' && 'type' in typeInfo) {
    return typeInfo.type.name
  } else {
    return typeInfo
  }
}

const convertAttribute = (
  attribute: Attr,
  propsTypes: any
): { name: string; value: any } => {
  let propName = toCamelCase(attribute.name)
  let propValue: string | number | boolean | null = attribute.value

  switch (getPropType(propsTypes[propName])) {
    case 'string':
    case String:
      propValue =
        ['null', 'undefined'].indexOf(propValue) >= 0 ? null : propValue
      break
    case 'number':
    case Number:
      propValue = Number(propValue)
      break
    case 'bool':
    case Boolean:
      propValue = !/^(false|0)$/i.test(propValue)
      break
    case 'object':
    case 'array':
    case Array:
    case Object:
      propValue = JSON.parse(propValue)
      break
    case 'func':
    case Function:
      propValue = eval(propValue)
      break
  }

  return {
    name: propName,
    value: propValue,
  }
}

interface RegisterOptions {
  modeShadow: boolean
  returnElement: boolean
  hasReactRef: boolean
  className: string
  lazy: boolean
}

type LazyReactComponentPromise = Promise<React.ComponentType<any>>
type LazyReactComponentType = () => LazyReactComponentPromise

type ReactComponentType = React.ComponentType<any> | LazyReactComponentType

type ReactComponent = ReactComponentType & {
  componentProps?: Record<string, any>
  propTypes?: Record<string, any>
}

interface IWebComponent extends HTMLElement {
  resolveProps: (props: any) => any
}

export function register(
  ReactComponent: ReactComponent,
  name: string,
  React: any,
  ReactDOM: any,
  options: Partial<RegisterOptions> = {}
): typeof HTMLElement | void {
  options = {
    modeShadow: false,
    returnElement: false,
    hasReactRef: false,
    className: 'html-element',
    lazy: false,
    ...options,
  }

  if (!React || !ReactDOM) {
    throw 'React and ReactDOM parameters must not be empty'
  }

  const getPropsTypes = (): any =>
    ReactComponent.componentProps || parsePropTypes(ReactComponent) || {}
  let propsTypes = getPropsTypes()

  class WebComponent extends HTMLElement implements IWebComponent {
    reactRoot: Root | null = null
    reactSlots: any = null
    reactElement: ReactElement | null = null
    reactChildren: ReactNode[] | null = null
    observer: MutationObserver | null = null

    connectedCallback() {
      this.setCustomId()
      this.setChildrenParent()

      this.observer = new MutationObserver(() => {
        if (this.isRendered()) {
          this.setIsHydrated()
          this.observer!.disconnect()

          const parent = this.getCustomParent()

          if (parent && parent.allChildrenHydrated()) {
            parent.parseChildAndRender()
          }
        }
      })

      this.observer.observe(this, { childList: true })

      if (this.allChildrenHydrated() && !this.isHydrated()) {
        this.parseChildAndRender()
      }
    }

    resolveProps(){
      resolveProps(propsTypes)
    }

    allChildrenHydrated(): boolean {
      return this.getChildrenNotHydrated().length === 0
    }

    setChildrenParent() {
      this.getChildrenNotHydrated().forEach((child) => {
        child.setAttribute('custom-parent', this.getAttribute('custom') ?? '')
      })
    }

    getChildrenNotHydrated(): NodeListOf<HTMLElement> {
      return this.querySelectorAll('[custom]:not([custom-state="hydrated"])')
    }

    getCustomParent(): WebComponent | null {
      const id = this.getAttribute('custom-parent')
      return this.closest(`[custom="${id}"]`)
    }

    setIsHydrated() {
      this.setAttribute('custom-state', 'hydrated')
    }

    setIsRendered() {
      this.setAttribute('custom-state', 'rendered')
    }

    isHydrated(): boolean {
      return this.getAttribute('custom-state') === 'hydrated'
    }

    isRendered(): boolean {
      return this.getAttribute('custom-state') === 'rendered'
    }

    setCustomId() {
      if (!Number.parseInt(this.getAttribute('custom') ?? '')) {
        this.setAttribute('custom', Math.random().toString(36))
      }
    }

    parseChildAndRender() {
      this.reactChildren = parseChildren(this.innerHTML)

      if (!options.lazy) {
        this.renderRoot()
      } else {
        let promise: LazyReactComponentPromise

        if (typeof ReactComponent === 'function') {
          promise = (ReactComponent as LazyReactComponentType)()
        } else {
          promise = ReactComponent
        }

        promise.then((resolvedComponent) => {
          ReactComponent = resolvedComponent
          propsTypes = getPropsTypes()
          this.resolveProps()
          this.renderRoot()
        })
      }
    }

    renderRoot() {
      if (!this.isHydrated()) {
        this.setIsRendered()
      }

      let props = { ...this.getProps(), ...this.getSlots() }

      if (options.hasReactRef) {
        props.ref = React.createRef() as RefObject<any>
      }

      props.rootElement = this

      if (!this.reactRoot) {
        let container = options.modeShadow
          ? this.attachShadow({ mode: 'open' })
          : this

        if (options.className) {
          ;(container as this).classList.add(options.className)
        }

        this.reactRoot = ReactDOM.createRoot(container)
      }

      this.reactElement = React.createElement(
        ReactComponent,
        props,
        this.reactChildren
      )

      this.reactRoot?.render(this.reactElement)
    }

    getProps(): any {
      return [...this.attributes]
        .filter((attr) => toCamelCase(attr.name) in propsTypes)
        .map((attr) => convertAttribute(attr, propsTypes))
        .reduce((props, prop) => ({ ...props, [prop.name]: prop.value }), {})
    }

    getSlots(): any {
      if (this.reactSlots) {
        return this.reactSlots
      }

      let slots: Record<string, any> = {}

      if (this.reactChildren instanceof Array) {
        this.reactChildren = this.reactChildren.filter((child) => {
          let name

          if (
            child &&
            typeof child !== 'string' &&
            typeof child !== 'number' &&
            typeof child !== 'boolean'
          ) {
            const anyChild = child as any
            if (anyChild?.type === 'slot' && anyChild.props?.name) {
              name = anyChild?.props.name
            } else if (anyChild?.props?.slot) {
              name = anyChild.props.slot
            }
          }

          const type = getPropType(propsTypes[name])
          if (name && (type === Node || type === 'node')) {
            slots[name] = child
            return false
          }

          return true
        })
      }

      if (Object.keys(slots).length > 0) {
        this.reactSlots = slots
      }

      return slots
    }

    attributeChangedCallback(name: string, oldValue: string, newValue: string) {
      if (this.reactRoot) {
        this.renderRoot()
      }
    }

    getReactRef(): any {
      if ((this.reactElement as any)?.ref?.current) {
        return (this.reactElement as any).ref.current
      }
      this.warnRefProblem()
    }

    getAsyncReactRef(): Promise<any> {
      let time = Date.now()
      return new Promise((resolve, reject) => {
        let interval = setInterval(() => {
          if ((this.reactElement as any)?.ref?.current) {
            clearInterval(interval)
            resolve((this.reactElement as any).ref.current)
          }

          if (time + 3000 < Date.now()) {
            this.warnRefProblem()
            clearInterval(interval)
            reject(false)
          }
        }, 0)
      })
    }

    warnRefProblem() {
      console.warn(
        'Ref not available, possible causes are:\n' +
          '- The component is not ready yet, use the getAsyncReactRef function instead.\n' +
          '- Check that the {hasReactRef: true} option has been added to the component registration\n' +
          '- Check that you have wrapped the React component with the state forwardRef\n\n' +
          'See the documentation for more info https://github.com/bougassaa/react-to-html-element#usage-of-ref'
      )
    }

    static get observedAttributes(): string[] {
      let propsNames = Object.keys(propsTypes)
      return [
        ...new Set([
          ...propsNames,
          ...propsNames.map((prop) => toDashedStyle(prop)),
        ]),
      ]
    }
  }

  function resolveProps(propsTypes: any) {
    Object.keys(propsTypes).forEach((propName) => {
      Object.defineProperty(WebComponent.prototype, propName, {
        get() {
          return this.getAttribute(toDashedStyle(propName))
        },
        set(value) {
          const v = typeof value === 'object' ? JSON.stringify(value) : value
          this.setAttribute(toDashedStyle(propName), v)
        },
      })
    })
  }

  if (!options.lazy) {
    resolveProps(propsTypes)
  }

  if (options.returnElement) {
    return WebComponent
  }

  customElements.define(name, WebComponent)
}
